! ===========================================================================
MODULE actlite_3yr_like

! ===========================================================================

  character(len=*), parameter,public :: ACT_data_dir = 'data/'
  character(len=*), parameter, public :: act_likelihood_version='ACTlite_3yr_v1'

  integer,public :: tt_lmax                = 3750  
  integer, parameter,public :: nbin       = 42  !max nbins in ACT data	
 
  !-------------------------------------------------------
  real(8) :: sig_e            =  0.04d0  !AR1 fractional error in power
  real(8) :: sig_s            =  0.04d0  
  real(8), parameter    :: PI    = 3.14159265358979323846264d0
  REAL(8) ::  btt_dat(nbin),btt_sig(nbin)
  REAL(8) ::  covmat(nbin,nbin),fisher(nbin,nbin),bval(nbin)
  REAL(8), dimension(:,:), allocatable :: win_func_e, win_func_s,win_func
  integer:: lmax_win           = 10000 !ell max of the full window functions  
  integer:: bmax0              = 34   !number of bins in full window function
  integer, parameter :: nbin_e   = 21  !bins in equa and south vector
  integer, parameter :: b1 = 5   !First bin of full window function used

  PRIVATE
  public :: act_likelihood_init
  public :: act_likelihood_compute
  public :: get_free_lun
contains
  
  ! ===========================================================================
  SUBROUTINE act_likelihood_init
    ! ===========================================================================
    
    IMPLICIT NONE
    
    INTEGER  :: i,j,lun,il,info
    CHARACTER(LEN=240) :: ttfilename, bblefilename, bblsfilename,covfilename
    LOGICAL  :: good
    
    print *, 'Initializing ACT likelihood, version '//act_likelihood_version
    
    !-----------------------------------------------
    ! set file names
    !-----------------------------------------------
    
    ttfilename  = trim(ACT_data_dir)//'ACT_cl_3yr_v1.dat'
    covfilename = trim(ACT_data_dir)//'ACT_cov_3yr_v1.dat'
    bblsfilename = trim(ACT_data_dir)//'Bbl_148_south_v1.dat'
    bblefilename = trim(ACT_data_dir)//'Bbl_148_equa_v1.dat'

    !-----------------------------------------------
    ! load spectrum and band power window functions 
    !-----------------------------------------------
    
    inquire(file=ttfilename,exist = good)
    if(.not.good)then
       write(*,*) 'cant find', trim(ttfilename), trim(ACT_data_dir)
       stop
    endif
    call get_free_lun( lun )
    open(unit=lun,file=ttfilename,form='formatted',status='unknown',action='read')
    
    do i=1,nbin
       read(lun,*) bval(i),btt_dat(i), btt_sig(i)
       btt_dat(i)=btt_dat(i)*(2.d0*PI)/(bval(i)*(bval(i)+1.d0)) !Convert back to Cl
       btt_sig(i)=btt_sig(i)*(2.d0*PI)/(bval(i)*(bval(i)+1.d0)) !Convert back to Cl
    enddo
    close(lun)
    
    call get_free_lun( lun )
    open(unit=lun,file=covfilename,form='formatted',status='unknown',action='read')
    
    do il = 1,nbin
       read(lun,*) (covmat(i,il),i=1,nbin)
       end do
    close(lun)

    inquire (file=bblefilename,exist = good)
    if(.not.good)then
       write(*,*) 'cant find', trim(bblefilename), trim(ACT_data_dir)
       stop
    endif
    call get_free_lun( lun )
    open(unit=lun,file=bblefilename,form='formatted',status='unknown',action='read')
    
    allocate(win_func_e(1:bmax0,1:lmax_win)) !Defined over ACT's full ell range 
    do il = 2, lmax_win
	read(lun,*) i, (win_func_e(i,il), i=1,bmax0) 
	enddo	
    close(lun)

    inquire (file=bblsfilename,exist = good)
    if(.not.good)then
       write(*,*) 'cant find', trim(bblsfilename), trim(ACT_data_dir)
       stop
    endif
    call get_free_lun( lun )
    open(unit=lun,file=bblsfilename,form='formatted',status='unknown',action='read')
    
    allocate(win_func_s(1:bmax0,1:lmax_win))
    do il = 2, lmax_win
       read(lun,*) i, (win_func_s(i,il), i=1,bmax0) 
    enddo
    close(lun)

!Combine the two windows into 1 array - ACT-E followed by ACT-S
    allocate(win_func(1:nbin,2:lmax_win)) 
    do i=1,nbin_e
       win_func(i,2:lmax_win)=win_func_e(i+b1-1,2:lmax_win) 
       win_func(i+nbin_e,2:lmax_win)=win_func_s(i+b1-1,2:lmax_win)
    enddo
    
  END SUBROUTINE act_likelihood_init
  
  ! ===========================================================================
  
  SUBROUTINE act_likelihood_compute(cltt,like)
    
    ! ===========================================================================
    
    IMPLICIT NONE
    REAL(8), intent(in) :: cltt(2:*)
    REAL(8), intent(out) :: like
    INTEGER :: lun,il,i,j,info
    REAL(8) :: cltt_temp(2:lmax_win)
    REAL(8) :: btt_th(nbin),btt_dum(nbin),diffs(nbin)
    REAL(8) :: tmp(nbin), tmp_e(nbin,1),tmp_s(nbin,1)
    REAL(8) :: dlnlike,lndet
    
    Like = 0.d0

    cltt_temp(2:lmax_win)=0.d0  !Neglect theory Cl et l>3750
    cltt_temp(2:tt_lmax)=cltt(2:tt_lmax)

!Convert theory back to Cl
    do il=2,tt_lmax
       cltt_temp(il)=cltt_temp(il)*2.d0*PI/(dble(il)*(dble(il)+1.d0)) 
    enddo
 
!Multiply theory Cls by bandpower matrix
    btt_th(1:nbin)=MATMUL(win_func(1:nbin,2:lmax_win),cltt_temp(2:lmax_win))

    !Basic chisq 
    diffs = btt_dat - btt_th

    !Replace cov matrix with one inflated for calibration error (could put in quick form to avoid re-inverting at every step, but not done here)
    tmp_e(:,:)=0.d0
    tmp_s(:,:)=0.d0
    tmp_e(1:nbin_e,1)=btt_th(1:nbin_e) !ACT-E theory
    tmp_s(nbin_e+1:nbin,1)=btt_th(nbin_e+1:nbin) !ACT-S theory
    fisher(:,:) = covmat(:,:)+sig_e**2*matmul(tmp_e,transpose(tmp_e))+sig_s**2*matmul(tmp_s,transpose(tmp_s))
    
    !Invert covmat
    call dpotrf('U',nbin,fisher,nbin,info)
    if(info.ne.0)then
       print*, ' info in dpotrf =', info
       stop
    endif
    call dpotri('U',nbin,fisher,nbin,info)
    if(info.ne.0)then
       print*, ' info in dpotri =', info
       stop
    endif
    do i=1,nbin
       do j=i,nbin
          fisher(j,i)=fisher(i,j)
       enddo
    enddo

    tmp=matmul(fisher,diffs)
    dlnlike=sum(tmp*diffs)

!Leave out lndet term as variation is <0.1 over calibration range considered
  !  lndet=0.d0
  !  do i=1,nbin
  !     lndet=lndet-log(fisher(i,i))
  !  enddo
  ! lndet=lndet/2.d0
  ! dlnlike=dlnlike+lndet
 
   Like = dlnlike/2.d0
    
  end SUBROUTINE act_likelihood_compute
  
 subroutine get_free_lun( lun )

    implicit none
    integer, intent(out) :: lun
    
    integer, save :: last_lun = 19
    logical :: used
    lun = last_lun
    do
       inquire( unit=lun, opened=used )
       if ( .not. used ) exit
       lun = lun + 1
    end do
    
    last_lun = lun
  end subroutine get_free_lun


END MODULE actlite_3yr_like

